"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var child_process_1 = require("child_process");
var rxjs_1 = require("rxjs");
var BlameParser = /** @class */ (function () {
    function BlameParser() {
        this.commitData = {};
        this.lineData = {};
        this.settingCommitData = false;
        this.currentCommitHash = '';
        this.currentLineNumber = 1;
    }
    BlameParser.prototype.parse = function (blame) {
        // Split up the original document into an array of lines
        var lines = blame.split('\n');
        // Go through each line
        // tslint:disable-next-line:prefer-for-of
        for (var _i = 0, lines_1 = lines; _i < lines_1.length; _i++) {
            var line = lines_1[_i];
            // If we detect a tab character we know it's a line of code
            // So we can reset stateful variables
            if (line[0] === '\t') {
                // The first tab is an addition made by git, so get rid of it
                this.lineData[this.currentLineNumber].code = line.substr(1);
                this.settingCommitData = false;
                this.currentCommitHash = '';
            }
            else {
                // If we are in the process of collecting data about a commit summary
                if (this.settingCommitData) {
                    this.parseCommitLine(line);
                }
                else {
                    var _a = line.split(' '), sha1 = _a[0], originalLineStr = _a[1], finalLineStr = _a[2], numLinesStr = _a[3];
                    this.currentCommitHash = sha1;
                    this.currentLineNumber = ~~finalLineStr;
                    // Since the commit data (author, committer, summary, etc) only
                    // appear once in a porcelain output for every commit, we set
                    // it up once here and then expect that the next 8-11 lines of
                    // the file are dedicated to that data
                    if (!this.commitData[sha1]) {
                        this.settingCommitData = true;
                        this.commitData[sha1] = { sha1: sha1 };
                    }
                    // Setup the new lineData hash
                    this.lineData[this.currentLineNumber] = {
                        code: '',
                        originalLine: ~~originalLineStr,
                        finalLine: ~~finalLineStr,
                        numLines: numLinesStr ? ~~numLinesStr : -1,
                        commit: this.commitData[sha1]
                    };
                }
            }
        }
    };
    /**
     * Parses and sets data from a line following a commit header
     *
     * @param {array} lineArr The current line split by a space
     */
    BlameParser.prototype.parseCommitLine = function (line) {
        var currentCommitData = this.commitData[this.currentCommitHash];
        var spaceIndex = line.indexOf(' ');
        var key = line.slice(0, spaceIndex);
        var value = line.slice(spaceIndex + 1);
        switch (key) {
            case 'author':
                currentCommitData.author = value;
                break;
            case 'author-mail':
                currentCommitData.authorMail = value;
                break;
            case 'author-time':
                currentCommitData.authorTime = new Date(~~value * 1000);
                break;
            case 'author-tz':
                currentCommitData.authorTz = value;
                break;
            case 'committer':
                currentCommitData.committer = value;
                break;
            case 'committer-mail':
                currentCommitData.committerMail = value;
                break;
            case 'committer-time':
                currentCommitData.committerTime = new Date(~~value * 1000);
                break;
            case 'committer-tz':
                currentCommitData.committerTz = value;
                break;
            case 'summary':
                currentCommitData.summary = value;
                break;
            case 'filename':
                currentCommitData.filename = value;
                break;
            case 'previous':
                currentCommitData.previousHash = value;
                break;
        }
    };
    return BlameParser;
}());
var Blamer = /** @class */ (function () {
    function Blamer() {
        /** Map from absolute file path to blame result */
        this.blames = new Map();
    }
    Blamer.prototype.blame = function (file, line) {
        var blamesObservable = this.blames.get(file);
        if (!blamesObservable) {
            blamesObservable = this.blameFile(file);
            this.blames.set(file, blamesObservable);
        }
        return blamesObservable
            .mergeMap(function (blames) {
            var blame = blames[line];
            if (!blame) {
                // Linters can report an error on the line that contains EOL, but git blame can't blame it
                return [];
            }
            return [blame.commit];
        });
    };
    Blamer.prototype.blameFile = function (file) {
        var _this = this;
        var obs = new rxjs_1.Observable(function (observer) {
            var child = child_process_1.spawn('git', ['blame', '--porcelain', file]);
            child.on('error', function (err) { return observer.error(err); });
            child.on('exit', function (exitCode) {
                if (!exitCode) {
                    observer.complete();
                }
                else {
                    observer.error(new Error("git blame " + file + " exited with " + exitCode + " " + stderr));
                }
            });
            var stderr = '';
            if (child.stdout) {
                child.stdout.on('data', function (chunk) { return observer.next(chunk); });
            }
            if (child.stderr) {
                child.stderr.on('data', function (chunk) { return stderr += chunk; });
            }
            return function () { return child.kill(); };
        })
            .retryWhen(function (errors) {
            return errors
                .do(function (err) {
                if (err.code !== 'EAGAIN') {
                    throw err;
                }
            })
                .delay(500)
                .take(10)
                .map(function (err, i) {
                if (i === 10) {
                    throw err;
                }
            });
        })
            .reduce(function (buffer, chunk) { return buffer + chunk; }, '')
            .map(function (output) {
            var blameParser = new BlameParser();
            blameParser.parse(output);
            return blameParser.lineData;
        })
            .do(undefined, function (err) {
            _this.blames.delete(file);
        })
            .publishReplay()
            .refCount();
        this.blames.set(file, obs);
        return obs;
    };
    return Blamer;
}());
exports.Blamer = Blamer;
//# sourceMappingURL=git.js.map